# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os

from conftest import create_project

from idf_ci.cli import click_cli


class TestPytestPlugin:
    def test_skip_tests_with_apps_not_built(self, pytester, runner):
        assert runner.invoke(click_cli, ['build', 'init', '--path', pytester.path]).exit_code == 0
        assert runner.invoke(click_cli, ['test', 'init', '--path', pytester.path]).exit_code == 0

        create_project('app1', pytester.path)
        create_project('app2', pytester.path)
        create_project('app3', pytester.path)

        pytester.maketxtfile(
            app_info_mock="""
            {"app_dir": "app1", "target": "esp32", "config_name": "default", "build_dir": "build_esp32_default", "build_system": "cmake", "build_status": "build success"}
            {"app_dir": "app2", "target": "esp32", "config_name": "default", "build_dir": "build_esp32_default", "build_system": "cmake", "build_status": "build success"}
            {"app_dir": "app3", "target": "esp32", "config_name": "default", "build_dir": "build_esp32_default", "build_system": "cmake", "build_status": "skipped"}
            """  # noqa: E501
        )

        pytester.makepyfile("""
                import pytest

                @pytest.mark.parametrize('target', ['esp32'], indirect=True)
                @pytest.mark.parametrize('app_path', ['app1', 'app2', 'app3'], indirect=True)
                def test_skip_tests(dut):
                    assert True
            """)
        res = pytester.runpytest('--target', 'esp32', '--log-cli-level', 'DEBUG', '-s')
        res.assert_outcomes(errors=2)  # failed because of no real builds

    def test_env_markers(self, pytester, runner):
        assert runner.invoke(click_cli, ['test', 'init', '--path', pytester.path]).exit_code == 0
        pytester.makepyfile("""
                import pytest
                from idf_ci.idf_pytest import PytestCase

                @pytest.mark.parametrize('target', ['esp32'], indirect=True)
                def test_env_markers(dut):
                    assert PytestCase.KNOWN_ENV_MARKERS == {'foo', 'bar'}
            """)

        os.makedirs(pytester.path / 'build')
        pytester.makefile(
            '.ini',
            pytest="""
                [pytest]
                env_markers =
                    foo: foo
                    bar: bar
            """,
        )

        res = pytester.runpytest()
        res.assert_outcomes(passed=1)



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import sys

import pytest

from idf_ci.utils import remove_subfolders


@pytest.mark.parametrize(
    'paths,expected',
    [
        (['/a/b/c', '/a/b', '/a', '/a/b/c/d'], ['/a']),
        (['/b', '/a/b', '/a', '/c/d'], ['/a', '/b', '/c/d']),
    ],
)
@pytest.mark.skipif(sys.platform == 'win32', reason='Using Unix paths')
def test_remove_subfolders(paths, expected):
    assert remove_subfolders(paths) == expected



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
"""This file is used for generating the child pipeline for build jobs."""

import logging
import os
import typing as t

import yaml
from idf_build_apps import App
from jinja2 import Environment

from idf_ci.envs import GitlabEnvVars
from idf_ci.idf_pytest import GroupedPytestCases, get_pytest_cases
from idf_ci.scripts import get_all_apps
from idf_ci.settings import CiSettings

logger = logging.getLogger(__name__)


def dump_apps_to_txt(apps: t.List[App], output_file: str) -> None:
    """Dump a list of apps to a text file, one app per line."""
    with open(output_file, 'w') as fw:
        for app in apps:
            fw.write(app.model_dump_json() + '\n')


def build_child_pipeline(
    *,
    paths: t.Optional[t.List[str]] = None,
    modified_files: t.Optional[t.List[str]] = None,
    compare_manifest_sha_filepath: t.Optional[str] = None,
    yaml_output: t.Optional[str] = None,
) -> None:
    """Generate build child pipeline."""
    envs = GitlabEnvVars()
    settings = CiSettings()

    if compare_manifest_sha_filepath and not os.path.isfile(compare_manifest_sha_filepath):
        compare_manifest_sha_filepath = None

    if yaml_output is None:
        yaml_output = settings.gitlab.build_pipeline.yaml_filename

    # Check if we should run quick pipeline
    if envs.select_by_filter_expr:
        # we only build test related apps
        test_related_apps, _ = get_all_apps(
            paths=paths,
            marker_expr='not host_test',
            filter_expr=envs.select_by_filter_expr,
        )
        non_test_related_apps: t.List[App] = []
        dump_apps_to_txt(test_related_apps, settings.collected_test_related_apps_filepath)
    else:
        test_related_apps, non_test_related_apps = get_all_apps(
            paths=paths,
            modified_files=modified_files,
            marker_expr='not host_test',
            compare_manifest_sha_filepath=compare_manifest_sha_filepath,
        )
        dump_apps_to_txt(test_related_apps, settings.collected_test_related_apps_filepath)
        dump_apps_to_txt(non_test_related_apps, settings.collected_non_test_related_apps_filepath)

    apps_total = len(test_related_apps) + len(non_test_related_apps)
    parallel_count = apps_total // settings.gitlab.build_pipeline.runs_per_job + 1

    if not apps_total:
        logger.info('No apps found, generating fake_pass job to skip the entire build child pipeline')
        with open(yaml_output, 'w') as fw:
            yaml.safe_dump(
                {
                    'fake_pass': {
                        'tags': settings.gitlab.build_pipeline.job_tags,
                        'stage': 'build',
                        'script': [
                            'echo "No apps found, skipping build child pipeline"',
                        ],
                        'before_script': [],
                        'after_script': [],
                    }
                },
                fw,
            )
            return

    logger.info(
        'Found %d apps, %d test related apps, %d non-test related apps',
        apps_total,
        len(test_related_apps),
        len(non_test_related_apps),
    )
    logger.info('Parallel count: %d', parallel_count)

    job_template = Environment().from_string(settings.gitlab.build_pipeline.job_template_jinja)
    jobs_template = Environment().from_string(settings.gitlab.build_pipeline.jobs_jinja)
    yaml_template = Environment().from_string(settings.gitlab.build_pipeline.yaml_jinja)

    with open(yaml_output, 'w') as fw:
        fw.write(
            yaml_template.render(
                job_template=job_template.render(
                    settings=settings,
                ),
                jobs=jobs_template.render(
                    settings=settings,
                    parallel_count=parallel_count,
                ),
                settings=settings,
            )
        )

    logger.info('Build child pipeline generated successfully in %s', yaml_output)


def test_child_pipeline(
    yaml_output: str,
    *,
    cases: t.Optional[GroupedPytestCases] = None,
) -> None:
    """This function is used to generate the child pipeline for test jobs.

    Suppose the ci_build_artifacts_filepatterns is downloaded already

    .. note::

        parallel:matrix does not support array as value, we generate all jobs here

    Example output:

    .. code-block:: yaml

        .default_test_settings:
            script:
                - pytest ${nodes}

        esp32 - generic:
            extends:
                - .default_test_settings
            tags:
                - esp32
                - generic
            variables:
                nodes: "nodeid1 nodeid2"
    """
    settings = CiSettings()

    if yaml_output is None:
        yaml_output = settings.gitlab.test_pipeline.yaml_filename

    if cases is None:
        cases = GroupedPytestCases(get_pytest_cases())

    if not cases.grouped_cases:
        logger.info('No test cases found, generating fake_pass job to skip the entire test child pipeline')
        with open(yaml_output, 'w') as fw:
            yaml.safe_dump(
                {
                    'fake_pass': {
                        # on purpose, cause test tags are programmatically generated
                        'tags': settings.gitlab.build_pipeline.job_tags,
                        'stage': 'test',
                        'script': [
                            'echo "No test cases found, skipping test child pipeline"',
                        ],
                        'before_script': [],
                        'after_script': [],
                    }
                },
                fw,
            )
        return

    jobs = []
    for key, grouped_cases in cases.grouped_cases.items():
        jobs.append(
            {
                'name': f'{key.target_selector} - {key.env_selector}',
                'tags': sorted(key.runner_tags),
                'nodes': ' '.join([c.item.nodeid for c in grouped_cases]),
                'parallel_count': len(grouped_cases) // settings.gitlab.test_pipeline.runs_per_job + 1,
                **cases.additional_dict.get(key, {}),
            }
        )

    job_template = Environment().from_string(settings.gitlab.test_pipeline.job_template_jinja)
    jobs_template = Environment().from_string(settings.gitlab.test_pipeline.jobs_jinja)
    yaml_template = Environment().from_string(settings.gitlab.test_pipeline.yaml_jinja)

    with open(yaml_output, 'w') as fw:
        fw.write(
            yaml_template.render(
                default_template=job_template.render(
                    settings=settings,
                ),
                jobs=jobs_template.render(
                    jobs=jobs,
                    settings=settings,
                ),
                settings=settings,
            )
        )

    logger.info('Test child pipeline generated successfully in %s', yaml_output)



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import re

import pytest

from idf_ci.cli import click_cli
from idf_ci.settings import CiSettings


def test_default_component_mapping_regexes():
    expected_regexes = [
        '/components/(.+)/',
        '/common_components/(.+)/',
    ]
    assert CiSettings().component_mapping_regexes == expected_regexes


def test_default_component_ignored_file_extensions():
    expected_extensions = [
        '.md',
        '.rst',
        '.yaml',
        '.yml',
        '.py',
    ]
    assert CiSettings().component_ignored_file_extensions == expected_extensions


def test_get_modified_components():
    test_files = [
        'components/wifi/wifi.c',
        'components/bt/bt_main.c',
        'common_components/esp_common/test.c',
        'docs/example.md',  # should be ignored
        'random/file.txt',  # should not match any component
    ]

    expected_components = {'wifi', 'bt', 'esp_common'}
    assert CiSettings().get_modified_components(test_files) == expected_components


def test_ignored_file_extensions():
    test_files = [
        'components/wifi/README.md',
        'components/bt/docs.rst',
        'components/esp_common/config.yaml',
        'components/test/test.yml',
        'components/utils/util.py',
    ]

    assert CiSettings().get_modified_components(test_files) == set()


def test_extended_component_mapping_regexes():
    settings = CiSettings(
        extend_component_mapping_regexes=[
            '/custom/path/(.+)/',
        ]
    )

    test_files = [
        'custom/path/my_component/test.c',
        'components/wifi/wifi.c',
    ]

    expected_components = {'my_component', 'wifi'}
    assert settings.get_modified_components(test_files) == expected_components


def test_extended_ignored_extensions():
    settings = CiSettings(
        extend_component_ignored_file_extensions=[
            '.txt',
            '.json',
        ]
    )

    test_files = [
        'components/wifi/test.txt',
        'components/bt/config.json',
        'components/esp_common/main.c',
    ]

    expected_components = {'esp_common'}
    assert settings.get_modified_components(test_files) == expected_components


def test_all_component_mapping_regexes():
    patterns = CiSettings().all_component_mapping_regexes
    assert len(patterns) == 2

    test_path = '/components/test_component/test.c'
    for pattern in patterns:
        match = pattern.search(test_path)
        if '/components/(.+)/' in pattern.pattern:
            assert match is not None
            assert match.group(1) == 'test_component'


def test_component_mapping_with_absolute_paths():
    abs_path = os.path.abspath('components/wifi/wifi.c')
    components = CiSettings().get_modified_components([abs_path])
    assert components == {'wifi'}


def test_ci_config_file_option(tmp_path, runner):
    custom_config = tmp_path / 'custom_ci_config.toml'
    with open(custom_config, 'w') as f:
        f.write("""
extend_component_mapping_regexes = [
    '/custom/path/(.+)/'
]

component_ignored_file_extensions = [
    '.custom'
]
""")
    runner.invoke(click_cli, ['--config-file', custom_config, 'build', 'init'])
    runner.invoke(click_cli, ['--config-file', custom_config, 'test', 'init'])

    result = runner.invoke(click_cli, ['--config-file', custom_config, 'build', 'run'])
    assert result.exit_code == 0
    assert CiSettings.CONFIG_FILE_PATH == custom_config
    assert len(CiSettings().all_component_mapping_regexes) == 3  # default got 2

    CiSettings.CONFIG_FILE_PATH = None  # reset

    # Test with non-existent config file
    non_existent = os.path.join(tmp_path, 'non_existent.toml')
    result = runner.invoke(click_cli, ['--config-file', non_existent])
    assert result.exit_code == 2  # Click returns 2 for parameter validation errors
    assert re.search(r"Error: Invalid value for '-c' / '--config-file': File .* does not exist.", result.output)
    assert len(CiSettings().all_component_mapping_regexes) == 2  # default got 2


def test_ci_profile_not_specified(runner):
    original_config_path = CiSettings.CONFIG_FILE_PATH
    with runner.isolated_filesystem() as tmp_d:
        result = runner.invoke(click_cli, ['build', 'init'])
        assert result.exit_code == 0
        assert CiSettings.CONFIG_FILE_PATH == original_config_path
        assert os.path.exists(tmp_d + os.sep + '.idf_build_apps.toml')


class TestEnvVars:
    @pytest.fixture(autouse=True)
    def _cleanup_envs(self):
        yield
        for key in list(os.environ.keys()):
            if key.startswith('TEST_VAR_'):
                del os.environ[key]

    def test_local_runtime_set_by_ci_settings(self, monkeypatch):
        CiSettings(
            local_runtime_envs={
                'TEST_VAR_BOOL': True,  # bool, but env var is always str
                'TEST_VAR_STR': 'test_local_value',
            },
            ci_runtime_envs={
                'TEST_VAR_BOOL': False,
                'TEST_VAR_STR': 'test_ci_value',
            },
            ci_detection_envs=['NON_EXIST_CI'],
        )
        assert os.getenv('TEST_VAR_BOOL') == 'True'
        assert os.getenv('TEST_VAR_STR') == 'test_local_value'

        monkeypatch.setenv('NON_EXIST_CI', '1')
        CiSettings(
            local_runtime_envs={
                'TEST_VAR_BOOL': True,  # bool, but env var is always str
                'TEST_VAR_STR': 'test_local_value',
            },
            ci_runtime_envs={
                'TEST_VAR_BOOL': False,
                'TEST_VAR_STR': 'test_ci_value',
            },
            ci_detection_envs=['NON_EXIST_CI'],
        )
        assert os.getenv('TEST_VAR_BOOL') == 'False'
        assert os.getenv('TEST_VAR_STR') == 'test_ci_value'

    def test_set_by_ci_config_file(self, tmp_path, monkeypatch):
        with open(tmp_path / '.idf_ci.toml', 'w') as f:
            f.write("""
ci_detection_envs=[
    'NON_EXIST_CI'
]

[local_runtime_envs]
TEST_VAR_BOOL = true  # bool, but env var is always str
TEST_VAR_STR = "test_local_value"

[ci_runtime_envs]
TEST_VAR_BOOL = false
TEST_VAR_STR = "test_ci_value"
""")

        CiSettings()
        assert os.getenv('TEST_VAR_BOOL') == 'True'
        assert os.getenv('TEST_VAR_STR') == 'test_local_value'

        monkeypatch.setenv('NON_EXIST_CI', '1')
        CiSettings()
        assert os.getenv('TEST_VAR_BOOL') == 'False'
        assert os.getenv('TEST_VAR_STR') == 'test_ci_value'



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os

import pytest

from idf_ci.idf_gitlab.scripts import pipeline_variables


@pytest.mark.parametrize(
    'env_vars,expected',
    [
        # Non-MR pipeline case
        (
            {'CI_COMMIT_SHA': '12345abcde'},
            {
                'IDF_CI_SELECT_ALL_PYTEST_CASES': '1',
                'PIPELINE_COMMIT_SHA': '12345abcde',
            },
        ),
        # MR pipeline with python constraint branch
        (
            {
                'CI_MERGE_REQUEST_IID': '123',
                'CI_COMMIT_SHA': 'bcdefa54321',
                'CI_MERGE_REQUEST_SOURCE_BRANCH_SHA': 'abcdef12345',
                'CI_PYTHON_CONSTRAINT_BRANCH': 'some-branch',
            },
            {
                'IDF_CI_SELECT_ALL_PYTEST_CASES': '1',
                'PIPELINE_COMMIT_SHA': 'abcdef12345',
            },
        ),
        # MR pipeline with BUILD_AND_TEST_ALL_APPS label
        (
            {
                'CI_MERGE_REQUEST_IID': '123',
                'CI_COMMIT_SHA': 'bcdefa54321',
                'CI_MERGE_REQUEST_SOURCE_BRANCH_SHA': 'abcdef12345',
                'CI_MERGE_REQUEST_LABELS': 'BUILD_AND_TEST_ALL_APPS,some-other-label',
            },
            {
                'IDF_CI_SELECT_ALL_PYTEST_CASES': '1',
                'PIPELINE_COMMIT_SHA': 'abcdef12345',
            },
        ),
        # MR pipeline with Test Case Filters in description
        (
            {
                'CI_MERGE_REQUEST_IID': '123',
                'CI_COMMIT_SHA': 'bcdefa54321',
                'CI_MERGE_REQUEST_SOURCE_BRANCH_SHA': 'abcdef12345',
                'CI_MERGE_REQUEST_DESCRIPTION': """
## Dynamic Pipeline Configuration

```yaml
Test Case Filters:
  - filter1
  - filter2
```

Some other text
""",
            },
            {
                'IDF_CI_SELECT_BY_FILTER_EXPR': 'filter1 or filter2',
                'IDF_CI_IS_DEBUG_PIPELINE': '1',
                'PIPELINE_COMMIT_SHA': 'abcdef12345',
            },
        ),
        # MR pipeline with no special conditions
        (
            {
                'CI_MERGE_REQUEST_IID': '123',
                'CI_COMMIT_SHA': 'bcdefa54321',
                'CI_MERGE_REQUEST_SOURCE_BRANCH_SHA': 'abcdef12345',
            },
            {
                'PIPELINE_COMMIT_SHA': 'abcdef12345',
            },
        ),
    ],
)
def test_pipeline_variables(monkeypatch, env_vars, expected):
    for env_var in [var for var in os.environ if var.startswith(('CI_', 'IDF_CI_'))]:
        monkeypatch.delenv(env_var, raising=False)

    for key, value in env_vars.items():
        monkeypatch.setenv(key, value)

    result = pipeline_variables()
    assert result == expected


def test_pipeline_variables_no_env_vars(monkeypatch):
    for env_var in [var for var in os.environ if var.startswith(('CI_', 'IDF_CI_'))]:
        monkeypatch.delenv(env_var, raising=False)

    result = pipeline_variables()
    assert result == {'IDF_CI_SELECT_ALL_PYTEST_CASES': '1'}



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import glob
import json
import logging
import os
import re
import subprocess
import typing as t
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from datetime import timedelta
from functools import lru_cache
from pathlib import Path

import esp_bool_parser
import minio
import requests
import urllib3
from gitlab import Gitlab
from minio import Minio

from .._compat import UNDEF, is_undefined
from .._vendor import translate
from ..envs import GitlabEnvVars
from ..settings import CiSettings
from ..utils import get_current_branch

logger = logging.getLogger(__name__)


def execute_concurrent_tasks(
    tasks: t.List[t.Callable[..., t.Any]],
    max_workers: t.Optional[int] = None,
    task_name: str = 'executing task',
) -> t.List[t.Any]:
    """Execute tasks concurrently using ThreadPoolExecutor.

    :param tasks: List of callable tasks to execute
    :param max_workers: Maximum number of worker threads
    :param task_name: Error message prefix for logging

    :returns: List of successful task results, sequence is not guaranteed
    """
    results = []
    errors = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(task) for task in tasks]

        for future in as_completed(futures):
            try:
                result = future.result()
                if result is not None:
                    results.append(result)
            except Exception as e:
                logger.error(f'Error while {task_name}: {e}')
                errors.append(e)

    if errors:
        _nl = '\n'  # compatible with Python < 3.12
        raise ArtifactError(f'Got {len(errors)} errors while {task_name}:\n{_nl.join([f"- {e}" for e in errors])}')

    return results


@dataclass
class ArtifactParams:
    """Common parameters for artifacts operations.

    The commit SHA can be determined in the following order of precedence:

    1. Explicitly provided commit_sha parameter
    2. PIPELINE_COMMIT_SHA environment variable
    3. Latest commit from branch (where branch is determined by branch parameter or
       current git branch)
    """

    commit_sha: t.Optional[str] = None
    branch: t.Optional[str] = None
    folder: t.Optional[str] = None

    def __post_init__(self):
        if self.folder is None:
            self.folder = os.getcwd()
        self.from_path = Path(self.folder)

        # Get commit SHA with the following precedence:
        # 1. CLI provided commit_sha
        if self.commit_sha:
            return

        # 2. Environment variable PIPELINE_COMMIT_SHA
        if os.getenv('PIPELINE_COMMIT_SHA'):
            self.commit_sha = os.environ['PIPELINE_COMMIT_SHA']
            return

        # 3. Latest commit from branch
        try:
            if self.branch is None:
                self.branch = get_current_branch()
            result = subprocess.run(
                ['git', 'rev-parse', self.branch],
                check=True,
                capture_output=True,
                encoding='utf-8',
            )
            self.commit_sha = result.stdout.strip()
        except Exception:
            raise ValueError(
                'Failed to get commit SHA from git command. '
                'Must set commit_sha or branch parameter, or set PIPELINE_COMMIT_SHA env var'
            )


class ArtifactError(RuntimeError):
    """Base exception for artifact-related errors."""


class S3Error(ArtifactError):
    """Exception raised for S3-related errors."""


class PresignedUrlError(ArtifactError):
    """Exception raised for presigned URL-related errors."""


class ArtifactManager:
    """Tool interface for managing artifacts in GitLab pipelines.

    This class provides a unified interface for downloading and uploading artifacts,
    supporting both GitLab's built-in storage and S3 storage. It handles:

    1. GitLab API operations (pipeline, merge request queries)
    2. S3 storage operations (artifact upload/download)
    3. Fallback to GitLab storage when S3 is not configured

    :var envs: GitLab environment variables
    :var settings: CI settings
    """

    def __init__(self):
        self.envs = GitlabEnvVars()
        self.settings = CiSettings()

        self._s3_client: t.Optional[Minio] = UNDEF  # type: ignore

    @property
    @lru_cache()
    def gl(self):
        return Gitlab(
            self.envs.GITLAB_HTTPS_SERVER,
            private_token=self.envs.GITLAB_ACCESS_TOKEN,
        )

    @property
    @lru_cache()
    def project(self):
        """Lazily initialize and cache the GitLab project."""
        project = self.gl.projects.get(self.settings.gitlab.project)
        if not project:
            raise ValueError(f'Project {self.settings.gitlab.project} not found')
        return project

    def _get_upload_details_by_type(self, artifact_type: t.Optional[str]) -> t.Dict[str, t.List[str]]:
        """Get file patterns grouped by bucket name based on the artifact type.

        :param artifact_type: Type of artifacts to download (debug, flash, metrics)

        :returns: Dictionary mapping bucket names to lists of patterns

        :raises ValueError: If the artifact type is invalid
        """
        _types = []
        if artifact_type:
            if artifact_type not in self.settings.gitlab.artifacts.available_s3_types:
                raise ValueError(
                    f'Invalid artifact type: {artifact_type}. '
                    f'Available types: {self.settings.gitlab.artifacts.available_s3_types}'
                )

            _types = [artifact_type]
        else:
            _types = self.settings.gitlab.artifacts.available_s3_types

        bucket_patterns: t.Dict[str, t.List[str]] = {}
        for artifact_type in _types:
            config = self.settings.gitlab.artifacts.s3[artifact_type]

            if config.get('if_clause'):
                try:
                    stmt = esp_bool_parser.parse_bool_expr(config['if_clause'])
                    res = stmt.get_value('', '')
                except Exception as e:
                    logger.info(
                        f'Skipping {artifact_type} artifacts due to error '
                        f'while evaluating if_clause: {config["if_clause"]}: {e}'
                    )
                    continue
                else:
                    if not res:
                        logger.info(f'Skipping {artifact_type} artifacts due to if_clause: {config["if_clause"]}')
                        continue

            bucket = config['bucket']
            if bucket not in bucket_patterns:
                bucket_patterns[bucket] = []
            bucket_patterns[bucket].extend(config['patterns'])

        if not bucket_patterns:
            logger.info('No S3 configured patterns found, skipping...')

        return bucket_patterns

    @property
    def s3_client(self) -> t.Optional[minio.Minio]:
        """Get or create the S3 client."""
        if is_undefined(self._s3_client):
            self._s3_client = self._create_s3_client()
        return self._s3_client

    def _create_s3_client(self) -> t.Optional[minio.Minio]:
        if not all(
            [
                self.envs.IDF_S3_SERVER,
                self.envs.IDF_S3_ACCESS_KEY,
                self.envs.IDF_S3_SECRET_KEY,
            ]
        ):
            logger.info('S3 credentials not available. Skipping S3 features...')
            return None

        if self.envs.IDF_S3_SERVER.startswith('https://'):
            host = self.envs.IDF_S3_SERVER.replace('https://', '')
            secure = True
        elif self.envs.IDF_S3_SERVER.startswith('http://'):
            host = self.envs.IDF_S3_SERVER.replace('http://', '')
            secure = False
        else:
            raise ValueError('Please provide a http or https server URL for S3')

        return minio.Minio(
            host,
            access_key=self.envs.IDF_S3_ACCESS_KEY,
            secret_key=self.envs.IDF_S3_SECRET_KEY,
            secure=secure,
            http_client=urllib3.PoolManager(
                num_pools=10,
                timeout=urllib3.Timeout.DEFAULT_TIMEOUT,
                retries=urllib3.Retry(
                    total=5,
                    backoff_factor=0.2,
                    status_forcelist=[500, 502, 503, 504],
                ),
            ),
        )

    def _get_s3_path(self, prefix: str, from_path: Path) -> str:
        if from_path.is_absolute():
            rel_path = str(from_path.relative_to(self.envs.IDF_PATH))
        else:
            rel_path = str(from_path)

        return f'{prefix}{rel_path}' if rel_path != '.' else prefix

    def _download_from_s3(
        self,
        s3_client: minio.Minio,
        *,
        bucket: str,
        prefix: str,
        from_path: Path,
        patterns: t.List[str],
    ) -> None:
        s3_path = self._get_s3_path(prefix, from_path)
        patterns_regexes = [re.compile(translate(pattern, recursive=True, include_hidden=True)) for pattern in patterns]

        def _download_task(_obj_name: str, _output_path: Path) -> None:
            logger.debug(f'Downloading {_obj_name} to {_output_path}')
            s3_client.fget_object(bucket, _obj_name, str(_output_path))

        tasks = []
        for obj in s3_client.list_objects(bucket, prefix=s3_path, recursive=True):
            output_path = Path(self.envs.IDF_PATH) / obj.object_name.replace(prefix, '')
            if not any(pattern.match(str(output_path)) for pattern in patterns_regexes):
                continue
            tasks.append(
                lambda _obj_name=obj.object_name, _output_path=output_path: _download_task(_obj_name, _output_path)
            )

        execute_concurrent_tasks(tasks, task_name='downloading object')

    def _upload_to_s3(
        self,
        s3_client: minio.Minio,
        *,
        bucket: str,
        prefix: str,
        from_path: Path,
        patterns: t.List[str],
    ) -> None:
        def _upload_task(_filepath: Path, _s3_path: str) -> None:
            logger.debug(f'Uploading {_filepath} to {_s3_path}')
            s3_client.fput_object(bucket, _s3_path, str(_filepath))

        tasks = []
        for pattern in patterns:
            abs_pattern = os.path.join(str(from_path), pattern)
            for file_str in glob.glob(abs_pattern, recursive=True):
                filepath = Path(file_str)
                if not filepath.is_file():
                    continue

                s3_path = self._get_s3_path(prefix, filepath)
                tasks.append(lambda _filepath=filepath, _s3_path=s3_path: _upload_task(_filepath, _s3_path))

        execute_concurrent_tasks(tasks, task_name='uploading file')

    def _download_from_presigned_json(self, presigned_json: str, from_path: Path, patterns: t.List[str]) -> None:
        with open(presigned_json) as f:
            presigned_urls = json.load(f)

        patterns_regexes = [re.compile(translate(pattern, recursive=True, include_hidden=True)) for pattern in patterns]

        for rel_path, url in presigned_urls.items():
            if from_path not in Path(rel_path).parents:
                continue

            output_path = Path(self.envs.IDF_PATH) / rel_path
            if not any(pattern.match(str(output_path)) for pattern in patterns_regexes):
                continue

            output_path.parent.mkdir(parents=True, exist_ok=True)

            response = requests.get(url, stream=True)
            if response.status_code != 200:
                raise PresignedUrlError(f'Failed to download {rel_path}: {response.status_code}')

            with open(output_path, 'wb') as f:
                f.write(response.content)
            logger.debug(f'Downloaded {rel_path} to {output_path}')

    def download_artifacts(
        self,
        *,
        commit_sha: t.Optional[str] = None,
        branch: t.Optional[str] = None,
        artifact_type: t.Optional[str] = None,
        folder: t.Optional[str] = None,
        presigned_json: t.Optional[str] = None,
    ) -> None:
        """Download artifacts from a pipeline.

        This method downloads artifacts from either GitLab's built-in storage or S3
        storage, depending on the configuration and artifact type.

        :param commit_sha: Optional commit SHA. If no commit_sha provided, will use 1)
            PIPELINE_COMMIT_SHA env var, 2) latest commit from branch
        :param branch: Optional Git branch. If no branch provided, will use current
            branch
        :param artifact_type: Type of artifacts to download (debug, flash, metrics)
        :param folder: download artifacts under this folder
        :param presigned_json: Path to the presigned.json file. If provided, will use
            this file to download artifacts. If not, will use s3 credentials to download
        """
        params = ArtifactParams(
            commit_sha=commit_sha,
            branch=branch,
            folder=folder,
        )

        if presigned_json:
            self._download_from_presigned_json(
                presigned_json,
                params.from_path,
                [p for patterns in self._get_upload_details_by_type(artifact_type).values() for p in patterns],
            )
            return

        if not self.s3_client:
            raise S3Error('Configure S3 storage to download artifacts')

        s3_prefix = f'{self.settings.gitlab.project}/{params.commit_sha}/'
        logger.info(f'Downloading artifacts under {params.from_path} from s3 (commit sha: {params.commit_sha})')

        for bucket, patterns in self._get_upload_details_by_type(artifact_type).items():
            self._download_from_s3(
                s3_client=self.s3_client,
                bucket=bucket,
                prefix=s3_prefix,
                from_path=params.from_path,
                patterns=patterns,
            )

    def upload_artifacts(
        self,
        *,
        commit_sha: t.Optional[str] = None,
        branch: t.Optional[str] = None,
        artifact_type: t.Optional[str] = None,
        folder: t.Optional[str] = None,
    ) -> None:
        """Upload artifacts to S3 storage.

        This method uploads artifacts to S3 storage only. GitLab's built-in storage is
        not supported. The commit SHA is required to identify where to store the
        artifacts.

        :param commit_sha: Optional commit SHA. If no commit_sha provided, will use 1)
            PIPELINE_COMMIT_SHA env var, 2) latest commit from branch
        :param branch: Optional Git branch. If no branch provided, will use current
            branch
        :param artifact_type: Type of artifacts to upload (debug, flash, metrics)
        :param folder: upload artifacts under this folder

        :raises S3Error: If S3 is not configured
        """
        params = ArtifactParams(
            commit_sha=commit_sha,
            branch=branch,
            folder=folder,
        )

        if not self.s3_client:
            raise S3Error('Configure S3 storage to upload artifacts')

        prefix = f'{self.settings.gitlab.project}/{params.commit_sha}/'
        logger.info(f'Uploading artifacts under {params.from_path} to s3 (commit sha: {params.commit_sha})')

        for bucket, patterns in self._get_upload_details_by_type(artifact_type).items():
            self._upload_to_s3(
                s3_client=self.s3_client,
                bucket=bucket,
                prefix=prefix,
                from_path=params.from_path,
                patterns=patterns,
            )

    def generate_presigned_json(
        self,
        *,
        commit_sha: t.Optional[str] = None,
        branch: t.Optional[str] = None,
        artifact_type: t.Optional[str] = None,
        folder: t.Optional[str] = None,
        expire_in_days: int = 4,
    ) -> t.Dict[str, str]:
        """Generate presigned URLs for artifacts in S3 storage.

        This method generates presigned URLs for artifacts that would be uploaded to S3
        storage. The URLs can be used to download the artifacts directly from S3.

        :param commit_sha: Optional commit SHA. If no commit_sha provided, will use 1)
            PIPELINE_COMMIT_SHA env var, 2) latest commit from branch
        :param branch: Optional Git branch. If no branch provided, will use current
            branch
        :param artifact_type: Type of artifacts to generate URLs for (debug, flash,
            metrics)
        :param folder: Base folder to generate relative paths from
        :param expire_in_days: Expiration time in days for the presigned URLs (default:
            4 days)

        :returns: Dictionary mapping relative paths to presigned URLs

        :raises S3Error: If S3 is not configured
        """
        params = ArtifactParams(
            commit_sha=commit_sha,
            branch=branch,
            folder=folder,
        )

        if not self.s3_client:
            raise S3Error('Configure S3 storage to generate presigned URLs')

        prefix = f'{self.settings.gitlab.project}/{params.commit_sha}/'
        s3_path = self._get_s3_path(prefix, params.from_path)

        def _get_presigned_url_task(_bucket: str, _obj_name: str) -> t.Tuple[str, str]:
            res = self.s3_client.get_presigned_url(  # type: ignore
                'GET',
                bucket_name=_bucket,
                object_name=_obj_name,
                expires=timedelta(days=expire_in_days),
            )
            if not res:
                raise S3Error(f'Failed to generate presigned URL for {_obj_name}')

            return _obj_name, res

        tasks = []
        presigned_urls: t.Dict[str, str] = {}
        for bucket, patterns in self._get_upload_details_by_type(artifact_type).items():
            patterns_regexes = [
                re.compile(translate(pattern, recursive=True, include_hidden=True)) for pattern in patterns
            ]

            for obj in self.s3_client.list_objects(bucket, prefix=s3_path, recursive=True):
                output_path = Path(self.envs.IDF_PATH) / obj.object_name.replace(prefix, '')
                if not any(pattern.match(str(output_path)) for pattern in patterns_regexes):
                    continue

                tasks.append(
                    lambda _bucket=bucket, _obj_name=obj.object_name: _get_presigned_url_task(_bucket, _obj_name)
                )

        results = execute_concurrent_tasks(tasks, task_name='generating presigned URL')
        for obj_name, presigned_url in results:
            presigned_urls[obj_name.replace(prefix, '')] = presigned_url

        return presigned_urls



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import logging
import subprocess
import typing as t
from pathlib import Path

from idf_build_apps.log import get_rich_log_handler

_T = t.TypeVar('_T')


@t.overload
def to_list(s: None) -> None: ...


@t.overload
def to_list(s: t.Iterable[_T]) -> t.List[_T]: ...


@t.overload
def to_list(s: _T) -> t.List[_T]: ...


def to_list(s):
    """Turn all objects to lists

    :param s: anything

    :returns: - ``None``, if ``s`` is None
        - itself, if ``s`` is a list
        - ``list(s)``, if ``s`` is a tuple or a set
        - ``[s]``, if ``s`` is other type
    """
    if s is None:
        return s

    if isinstance(s, list):
        return s

    if isinstance(s, set) or isinstance(s, tuple):
        return list(s)

    return [s]


def setup_logging(level: t.Optional[int] = logging.INFO) -> None:
    """Setup logging

    :param level: logging level
    """
    if level is None:
        level = logging.INFO

    package_logger = logging.getLogger(__package__)
    package_logger.setLevel(level)

    if package_logger.hasHandlers():
        package_logger.handlers.clear()
    package_logger.addHandler(get_rich_log_handler(level))

    package_logger.propagate = False


def remove_subfolders(paths: t.List[str]) -> t.List[str]:
    """Remove paths that are subfolders of other paths in the list.

    :param paths: List of directory paths as strings

    :returns: Filtered list of paths with no subfolder/parent folder relationships
    """
    result = set()

    for p in sorted([Path(p).resolve() for p in paths]):
        if not any(parent in result for parent in p.parents):
            result.add(p)

    return sorted([str(p) for p in result])


def get_current_branch() -> str:
    """Get the current Git branch by running git command.

    :returns: The current Git ref (branch name or commit SHA)

    :raises RuntimeError: If not in a Git repository or git command fails
    """
    result = subprocess.run(
        ['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
        check=True,
        capture_output=True,
        encoding='utf-8',
    )
    branch = result.stdout.strip()

    # If in detached HEAD state, rev-parse returns "HEAD"
    if branch == 'HEAD':
        raise RuntimeError('Failed to get current Git ref. Are you in a Git repository?')

    return branch



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import logging
import typing as t

from pydantic_settings import (
    BaseSettings,
)

logger = logging.getLogger(__name__)


class GitlabEnvVars(BaseSettings):
    # Pipeline Control Variables
    CHANGED_FILES_SEMICOLON_SEPARATED: t.Optional[str] = None
    """Semicolon-separated list of changed files in the pipeline."""

    # GitLab API Authentication
    GITLAB_HTTPS_SERVER: str = 'https://gitlab.com'
    """GitLab server URL for API calls."""

    GITLAB_ACCESS_TOKEN: t.Optional[str] = None
    """Access token for GitLab API authentication."""

    # S3 Storage Configuration
    IDF_S3_SERVER: t.Optional[str] = None
    """S3 server endpoint URL."""

    IDF_S3_ACCESS_KEY: t.Optional[str] = None
    """S3 access key for authentication."""

    IDF_S3_SECRET_KEY: t.Optional[str] = None
    """S3 secret key for authentication."""

    # other env vars used
    IDF_PATH: str = ''
    """Path to the ESP-IDF directory."""

    # Possibly Set by `idf-ci gitlab dynamic-pipeline-variables`
    IDF_CI_IS_DEBUG_PIPELINE: t.Optional[bool] = None
    """Flag indicating whether this is a debug pipeline."""

    IDF_CI_SELECT_BY_FILTER_EXPR: t.Optional[str] = None
    """Filter expression for selecting pytest cases."""

    IDF_CI_SELECT_ALL_PYTEST_CASES: t.Optional[bool] = None
    """Flag indicating whether to select all pytest cases."""

    @property
    def select_all_pytest_cases(self) -> bool:
        """Determine if all pytest cases should be selected.

        :returns: True if this is a full pipeline run, False otherwise
        """
        if self.IDF_CI_SELECT_ALL_PYTEST_CASES == '1':
            logger.info('Selecting all pytest cases since `IDF_CI_SELECT_ALL_PYTEST_CASES=1`')
            return True

        return False

    @property
    def select_by_filter_expr(self) -> t.Optional[str]:
        """Get the filter expression for pytest cases.

        :returns: The filter expression if set, None otherwise
        """
        if self.IDF_CI_SELECT_BY_FILTER_EXPR:
            logger.info('Selecting pytest cases with filter expression: %s', self.IDF_CI_SELECT_BY_FILTER_EXPR)
            return self.IDF_CI_SELECT_BY_FILTER_EXPR

        return None



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import fnmatch
import logging
import os
import typing as t
from dataclasses import dataclass

from idf_build_apps import App, build_apps, find_apps
from idf_build_apps.constants import SUPPORTED_TARGETS, BuildStatus
from idf_build_apps.utils import get_parallel_start_stop

from ._compat import UNDEF, UndefinedOr, is_defined_and_satisfies
from .envs import GitlabEnvVars
from .settings import CiSettings

logger = logging.getLogger(__name__)


@dataclass
class ProcessedArgs:
    """Container for processed arguments with meaningful field names."""

    modified_files: t.Optional[t.List[str]]
    modified_components: t.Optional[t.List[str]]
    filter_expr: t.Optional[str]
    default_build_targets: t.List[str]
    test_related_apps: t.Optional[t.List[App]]
    non_test_related_apps: t.Optional[t.List[App]]


def preprocess_args(
    modified_files: t.Optional[t.List[str]] = None,
    modified_components: t.Optional[t.List[str]] = None,
    filter_expr: t.Optional[str] = None,
    default_build_targets: t.Optional[t.List[str]] = None,
) -> ProcessedArgs:
    """Set values according to the environment variables, .toml settings, and defaults.

    :param modified_files: List of modified files
    :param modified_components: List of modified components
    :param filter_expr: Pytest filter expression
    :param default_build_targets: Default build targets to use

    :returns: Processed arguments as a ProcessedArgs object
    """
    envs = GitlabEnvVars()
    settings = CiSettings()

    processed_targets: t.List[str] = SUPPORTED_TARGETS if default_build_targets is None else default_build_targets
    if settings.extra_default_build_targets:
        processed_targets = [*processed_targets, *settings.extra_default_build_targets]

    if envs.select_all_pytest_cases:
        return ProcessedArgs(
            modified_files=None,
            modified_components=None,
            filter_expr=None,
            default_build_targets=processed_targets,
            test_related_apps=None,
            non_test_related_apps=None,
        )

    processed_files = modified_files
    if processed_files is None and is_defined_and_satisfies(envs.CHANGED_FILES_SEMICOLON_SEPARATED):
        processed_files = envs.CHANGED_FILES_SEMICOLON_SEPARATED.split(';')  # type: ignore

    processed_components = modified_components
    if processed_files is not None and processed_components is None:
        processed_components = sorted(settings.get_modified_components(processed_files))

    processed_filter = envs.IDF_CI_SELECT_BY_FILTER_EXPR if filter_expr is None else filter_expr
    if processed_filter is not None:
        logger.info(
            'Running with quick test filter: %s. Skipping dependency-driven build. Build and test only filtered cases.',
            processed_filter,
        )
        processed_files = None
        processed_components = None

    test_related_apps = settings.read_apps_from_files([settings.collected_test_related_apps_filepath])
    non_test_related_apps = settings.read_apps_from_files([settings.collected_non_test_related_apps_filepath])

    # if one of the two is None, it should be empty list
    if test_related_apps is None and non_test_related_apps is not None:
        test_related_apps = []
    elif test_related_apps is not None and non_test_related_apps is None:
        non_test_related_apps = []

    return ProcessedArgs(
        modified_files=processed_files,
        modified_components=processed_components,
        filter_expr=processed_filter,
        default_build_targets=processed_targets,
        test_related_apps=test_related_apps,
        non_test_related_apps=non_test_related_apps,
    )


def get_all_apps(
    *,
    paths: t.Optional[t.List[str]] = None,
    target: str = 'all',
    # args that may be set by env vars or .idf_ci.toml
    modified_files: t.Optional[t.List[str]] = None,
    modified_components: t.Optional[t.List[str]] = None,
    filter_expr: t.Optional[str] = None,
    default_build_targets: t.Optional[t.List[str]] = None,
    # args that may be set by target
    marker_expr: UndefinedOr[t.Optional[str]] = UNDEF,
    # additional args
    compare_manifest_sha_filepath: t.Optional[str] = None,
    build_system: str = 'cmake',
) -> t.Tuple[t.List[App], t.List[App]]:
    """Get test-related and non-test-related applications.

    :param paths: List of paths to search for applications
    :param target: Target device(s) separated by commas
    :param modified_files: List of modified files
    :param modified_components: List of modified components
    :param filter_expr: Pytest filter expression -k
    :param default_build_targets: Default build targets to use
    :param marker_expr: Pytest marker expression -m
    :param compare_manifest_sha_filepath: Path to the manifest SHA file generated by
        `idf-build-apps dump-manifest-sha`
    :param build_system: Filter the apps by build system. Can be "cmake", "make" or a
        custom App class path

    :returns: Tuple of (test_related_apps, non_test_related_apps)
    """
    settings = CiSettings()
    processed_args = preprocess_args(
        modified_files=modified_files,
        modified_components=modified_components,
        filter_expr=filter_expr,
        default_build_targets=default_build_targets,
    )

    if processed_args.test_related_apps is not None and processed_args.non_test_related_apps is not None:
        for app in processed_args.test_related_apps:
            app.preserve = settings.preserve_test_related_apps

        for app in processed_args.non_test_related_apps:
            app.preserve = settings.preserve_non_test_related_apps

        return processed_args.test_related_apps, processed_args.non_test_related_apps

    paths = paths or ['.']

    apps = []
    for _t in target.split(','):
        if _t != 'all' and _t not in processed_args.default_build_targets:
            _default_build_targets = [*processed_args.default_build_targets, _t]
        else:
            _default_build_targets = processed_args.default_build_targets

        apps.extend(
            find_apps(
                paths,
                _t,
                modified_files=processed_args.modified_files,
                modified_components=processed_args.modified_components,
                include_skipped_apps=True,
                default_build_targets=_default_build_targets,
                compare_manifest_sha_filepath=compare_manifest_sha_filepath,
                build_system=build_system,
            )
        )

    # avoid circular import
    from .idf_pytest import get_pytest_cases

    cases = get_pytest_cases(
        paths=paths, target=target, marker_expr=marker_expr, filter_expr=processed_args.filter_expr
    )
    if not cases:
        for app in apps:
            app.preserve = settings.preserve_non_test_related_apps
        return [], sorted(apps)

    # Get modified pytest cases if any
    modified_pytest_cases = []
    if processed_args.modified_files:
        modified_pytest_scripts = [
            os.path.dirname(f)
            for f in processed_args.modified_files
            if fnmatch.fnmatch(os.path.basename(f), 'test_*.py')
        ]
        if modified_pytest_scripts:
            modified_pytest_cases = get_pytest_cases(
                paths=modified_pytest_scripts,
                target=target,
                marker_expr=marker_expr,
                filter_expr=processed_args.filter_expr,
            )

    # Create dictionaries mapping app info to test cases
    def get_app_dict(_cases):
        return {(case_app.path, case_app.target, case_app.config): _case for _case in _cases for case_app in _case.apps}

    pytest_dict = get_app_dict(cases)
    modified_pytest_dict = get_app_dict(modified_pytest_cases)

    test_apps = set()
    non_test_apps = set()

    for app in apps:
        app_key = (os.path.abspath(app.app_dir), app.target, app.config_name or 'default')
        # override build_status if test script got modified
        case = modified_pytest_dict.get(app_key)
        if case:
            test_apps.add(app)
            app.build_status = BuildStatus.SHOULD_BE_BUILT
            logger.debug('Found app: %s - required by modified test case %s', app, case.path)
        elif app.build_status != BuildStatus.SKIPPED:
            case = pytest_dict.get(app_key)
            if case:
                test_apps.add(app)
                # build or not should be decided by the build stage
                logger.debug('Found test-related app: %s - required by %s', app, case.path)
            else:
                non_test_apps.add(app)
                logger.debug('Found non-test-related app: %s', app)

    for app in test_apps:
        app.preserve = settings.preserve_test_related_apps

    for app in non_test_apps:
        app.preserve = settings.preserve_non_test_related_apps

    return sorted(test_apps), sorted(non_test_apps)


def build(
    *,
    paths: t.Optional[t.List[str]] = None,
    target: str = 'all',
    parallel_count: int = 1,
    parallel_index: int = 1,
    modified_files: t.Optional[t.List[str]] = None,
    modified_components: t.Optional[t.List[str]] = None,
    only_test_related: bool = False,
    only_non_test_related: bool = False,
    dry_run: bool = False,
    build_system: str = 'cmake',
    marker_expr: UndefinedOr[str] = UNDEF,
    filter_expr: t.Optional[str] = None,
) -> t.Tuple[t.List[App], int]:
    """Build applications based on specified parameters.

    :param paths: List of paths to search for applications
    :param target: Target device(s) separated by commas
    :param parallel_count: Total number of parallel jobs
    :param parallel_index: Index of current parallel job (1-based)
    :param modified_files: List of modified files
    :param modified_components: List of modified components
    :param only_test_related: Only build test-related applications
    :param only_non_test_related: Only build non-test-related applications
    :param dry_run: Do not actually build, just simulate
    :param build_system: Filter the apps by build system. Can be "cmake", "make" or a
        custom App class path
    :param marker_expr: Pytest marker expression
    :param filter_expr: Filter expression

    :returns: Tuple of (built apps, build return code)
    """
    settings = CiSettings()

    # Preprocess arguments
    processed_args = preprocess_args(
        modified_files=modified_files,
        modified_components=modified_components,
        filter_expr=filter_expr,
    )

    test_related_apps, non_test_related_apps = get_all_apps(
        paths=paths,
        target=target,
        modified_files=processed_args.modified_files,
        modified_components=processed_args.modified_components,
        build_system=build_system,
        marker_expr=marker_expr,
        filter_expr=processed_args.filter_expr,
    )

    for app in test_related_apps:
        app.preserve = settings.preserve_test_related_apps

    for app in non_test_related_apps:
        app.preserve = settings.preserve_non_test_related_apps

    if not only_test_related and not only_non_test_related:
        apps = sorted([*test_related_apps, *non_test_related_apps])
    elif only_test_related:
        apps = test_related_apps
    else:
        apps = non_test_related_apps

    ret = build_apps(
        apps,
        parallel_count=parallel_count,
        parallel_index=parallel_index,
        dry_run=dry_run,
        modified_files=processed_args.modified_files,
        modified_components=processed_args.modified_components,
    )

    # only returning the ones assigned, 1-indexed
    start, stop = get_parallel_start_stop(len(apps), parallel_count, parallel_index)

    return apps[start - 1 : stop], ret



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import json

import click

from idf_ci.cli._options import (
    option_branch,
    option_commit_sha,
    option_modified_files,
    option_paths,
)
from idf_ci.idf_gitlab import ArtifactManager
from idf_ci.idf_gitlab import build_child_pipeline as build_child_pipeline_cmd
from idf_ci.idf_gitlab import pipeline_variables as pipeline_variables_cmd
from idf_ci.idf_gitlab import test_child_pipeline as test_child_pipeline_cmd
from idf_ci.settings import CiSettings

_SETTINGS = CiSettings()


@click.group()
def gitlab():
    """Group of gitlab related commands"""
    pass


@gitlab.command()
def pipeline_variables():
    """Output dynamic pipeline variables.

    Analyzes the current GitLab pipeline environment and determines what variables to
    set for controlling pipeline behavior. Outputs variables in the format KEY="VALUE"
    for each determined variable, which can be used with GitLab's `export` feature.

    As for the generated variables, please refer to the following link:

    \b
    https://docs.espressif.com/projects/idf-ci/en/latest/references/api/idf_ci.idf_gitlab.html#idf_ci.idf_gitlab.pipeline_variables
    """
    for k, v in pipeline_variables_cmd().items():
        click.echo(f'{k}={v}')


@gitlab.command()
@option_paths
@option_modified_files
@click.option(
    '--compare-manifest-sha-filepath',
    default='.manifest_sha',
    help='Path to the recorded manifest sha file generated by `idf-build-apps dump-manifest-sha`',
)
@click.argument('yaml_output', required=False)
def build_child_pipeline(paths, modified_files, compare_manifest_sha_filepath, yaml_output):
    """Generate build child pipeline yaml file."""
    build_child_pipeline_cmd(
        paths=paths,
        modified_files=modified_files,
        compare_manifest_sha_filepath=compare_manifest_sha_filepath,
        yaml_output=yaml_output,
    )


@gitlab.command()
@click.argument('yaml_output', required=False)
def test_child_pipeline(yaml_output):
    """Generate test child pipeline yaml file."""
    test_child_pipeline_cmd(yaml_output)


@gitlab.command()
@click.option(
    '--type',
    'artifact_type',
    type=click.Choice(_SETTINGS.gitlab.artifacts.available_s3_types),
    help='Type of artifacts to download. If not specified, downloads all types.',
)
@option_commit_sha
@option_branch
@click.option(
    '--presigned-json',
    type=click.Path(dir_okay=False, file_okay=True, exists=True),
    help='Path to the presigned.json file.',
)
@click.argument('folder', required=False)
def download_artifacts(artifact_type, commit_sha, branch, folder, presigned_json):
    """Download artifacts from a GitLab pipeline.

    This command downloads artifacts from either GitLab's built-in storage or S3
    storage, depending on the configuration. The artifacts are downloaded to the
    specified folder (or current directory if not specified).
    """
    manager = ArtifactManager()
    manager.download_artifacts(
        commit_sha=commit_sha,
        branch=branch,
        artifact_type=artifact_type,
        folder=folder,
        presigned_json=presigned_json,
    )


@gitlab.command()
@click.option(
    '--type',
    'artifact_type',
    type=click.Choice(_SETTINGS.gitlab.artifacts.available_s3_types),
    help='Type of artifacts to upload',
)
@option_commit_sha
@option_branch
@click.argument('folder', required=False)
def upload_artifacts(artifact_type, commit_sha, branch, folder):
    """Upload artifacts to S3 storage.

    This command uploads artifacts to S3 storage only. GitLab's built-in storage is not
    supported. The commit SHA is required to identify where to store the artifacts.
    """
    manager = ArtifactManager()
    manager.upload_artifacts(
        commit_sha=commit_sha,
        branch=branch,
        artifact_type=artifact_type,
        folder=folder,
    )


@gitlab.command()
@option_commit_sha
@option_branch
@click.option(
    '--type',
    'artifact_type',
    type=click.Choice(_SETTINGS.gitlab.artifacts.available_s3_types),
    help='Type of artifacts to generate presigned URLs for',
)
@click.option(
    '--expire-in-days',
    type=int,
    default=4,
    help='Expiration time in days for the presigned URLs (default: 4 days)',
)
@click.option(
    '-o',
    '--output',
    type=click.Path(dir_okay=False, file_okay=True),
    help='Path to save the generated presigned URLs. If not specified, will print to stdout.',
)
@click.argument('folder', required=False)
def generate_presigned_json(commit_sha, branch, artifact_type, expire_in_days, output, folder):
    """Generate presigned URLs for artifacts in S3 storage.

    This command generates presigned URLs for artifacts that would be uploaded to S3
    storage. The URLs can be used to download the artifacts directly from S3.
    """
    manager = ArtifactManager()
    presigned_urls = manager.generate_presigned_json(
        commit_sha=commit_sha,
        branch=branch,
        artifact_type=artifact_type,
        folder=folder,
        expire_in_days=expire_in_days,
    )

    if output:
        with open(output, 'w') as f:
            json.dump(presigned_urls, f)
    else:
        click.echo(json.dumps(presigned_urls))


@gitlab.command()
@click.argument('filename', required=True)
def download_known_failure_cases_file(filename):
    """Download known failure cases file from S3 storage."""
    s3_client = ArtifactManager().s3_client

    if s3_client:
        s3_client.fget_object(
            _SETTINGS.gitlab.known_failure_cases_bucket_name,
            filename,
            filename,
        )
    else:
        raise ValueError('Configure S3 storage to download artifacts')



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import time

import click

from idf_ci.scripts import build as build_cmd

from ._options import (
    create_config_file,
    option_modified_files,
    option_parallel,
    option_paths,
    option_pytest,
    option_target,
)


@click.group()
def build():
    """Group of build related commands"""
    pass


@build.command()
@option_paths
@option_target
@option_parallel
@option_pytest
@option_modified_files
@click.option('--only-test-related', is_flag=True, help='Run build only for test-related apps')
@click.option('--only-non-test-related', is_flag=True, help='Run build only for non-test-related apps')
@click.option('--dry-run', is_flag=True, help='Run build in dry-run mode')
@click.option(
    '--build-system',
    default='cmake',
    help='Filter the apps by build system. Can be "cmake", "make" or a custom App class path in format "module:class"',
)
@click.pass_context
def run(
    ctx,
    *,
    paths,
    target,
    parallel_count,
    parallel_index,
    modified_files,
    only_test_related,
    only_non_test_related,
    dry_run,
    build_system,
    marker_expr,
    filter_expr,
):
    """Execute the build process for applications"""
    start_time = time.time()
    apps, ret = build_cmd(
        paths=paths,
        target=target,
        parallel_count=parallel_count,
        parallel_index=parallel_index,
        modified_files=modified_files,
        only_test_related=only_test_related,
        only_non_test_related=only_non_test_related,
        dry_run=dry_run,
        build_system=build_system,
        marker_expr=marker_expr,
        filter_expr=filter_expr,
    )
    click.echo(f'Built the following apps in {time.time() - start_time:.2f} seconds:')
    for app in apps:
        line = f'\t{app.build_path} [{app.build_status.value}]'
        if app.build_comment:
            line += f' ({app.build_comment})'
        click.echo(line)

    ctx.exit(ret)


@build.command()
@click.option('--path', help='Path to create the config file')
def init(path: str):
    """Create .idf_build_apps.toml with default values"""
    create_config_file(os.path.join(os.path.dirname(__file__), '..', 'templates', '.idf_build_apps.toml'), path)



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import io
import logging
import os.path
import typing as t
from contextlib import redirect_stderr, redirect_stdout

import pytest
from _pytest.config import ExitCode

from idf_ci._compat import UNDEF, UndefinedOr, is_undefined
from idf_ci.envs import GitlabEnvVars

from ..utils import remove_subfolders, setup_logging
from .models import PytestCase
from .plugin import IdfPytestPlugin

logger = logging.getLogger(__name__)


def get_pytest_cases(
    *,
    paths: t.Optional[t.List[str]] = None,
    target: str = 'all',
    sdkconfig_name: t.Optional[str] = None,
    marker_expr: UndefinedOr[t.Optional[str]] = UNDEF,
    filter_expr: t.Optional[str] = None,
) -> t.List[PytestCase]:
    """Collect pytest test cases from specified paths.

    :param paths: List of file system paths to collect test cases from
    :param target: Filter by targets
    :param sdkconfig_name: Filter tests whose apps are built with this sdkconfig name
    :param marker_expr: Filter by pytest marker expression -m
    :param filter_expr: Filter by pytest filter expression -k

    :returns: List of collected PytestCase objects

    :raises RuntimeError: If pytest collection fails
    """
    envs = GitlabEnvVars()

    paths = paths or ['.']

    if is_undefined(marker_expr):
        marker_expr = 'host_test' if 'linux' in target else 'not host_test'

    if filter_expr is None:
        filter_expr = envs.IDF_CI_SELECT_BY_FILTER_EXPR

    plugin = IdfPytestPlugin(
        cli_target=target,
        sdkconfig_name=sdkconfig_name,
    )

    args = [
        # remove sub folders if parent folder is already in the list
        # https://github.com/pytest-dev/pytest/issues/13319
        *remove_subfolders(paths),
        '--collect-only',
        '--rootdir',
        os.getcwd(),
        '--target',
        target,
    ]

    if marker_expr:
        args.extend(['-m', f'{marker_expr}'])
    if filter_expr:
        args.extend(['-k', f'{filter_expr}'])

    logger.debug('Collecting pytest test cases with args: %s', args)

    original_log_level = logger.parent.level  # type: ignore
    with io.StringIO() as stdout_buffer, io.StringIO() as stderr_buffer:
        with redirect_stdout(stdout_buffer), redirect_stderr(stderr_buffer):
            result = pytest.main(args, plugins=[plugin])
        stdout_content = stdout_buffer.getvalue()
        stderr_content = stderr_buffer.getvalue()

    # Restore logging level as redirection changes it
    setup_logging(level=original_log_level)

    if result == ExitCode.OK:
        return plugin.cases

    raise RuntimeError(f'pytest collection failed.\nArgs: {args}\nStdout: {stdout_content}\nStderr: {stderr_content}')



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import os
import shutil
import subprocess
import sys
import textwrap

import minio
import pytest
import requests

from idf_ci.cli import click_cli
from idf_ci.idf_gitlab import ArtifactManager
from idf_ci.idf_gitlab.api import S3Error


# to run this test, don't forget to run "docker compose up -d" in the root directory of the project
@pytest.mark.skipif(sys.platform == 'win32', reason='minio service not available on Windows')
class TestUploadDownloadArtifacts:
    @pytest.fixture
    def sample_artifacts_dir(self, tmp_path):
        build_dir = tmp_path / 'app' / 'build_esp32_build'
        build_dir.mkdir(parents=True)

        # Create some test files
        (build_dir / 'build.log').write_text('Test build log', encoding='utf-8')
        (build_dir / 'test.bin').write_text('Binary content', encoding='utf-8')
        (build_dir / 'size.json').write_text('{"size": 1024}', encoding='utf-8')
        (build_dir / 'optional.txt').write_text('Optional content', encoding='utf-8')

        return build_dir

    @pytest.fixture(autouse=True)
    def setup_test_dir(self, tmp_path, tmp_dir, monkeypatch):
        (tmp_path / '.idf_ci.toml').write_text(
            textwrap.dedent("""
                [gitlab]
                project = "espressif/esp-idf"

                [gitlab.artifacts.s3.debug]
                bucket = "test-bucket"
                patterns = ["**/build*/build.log"]

                [gitlab.artifacts.s3.flash]
                bucket = "test-bucket"
                patterns = ["**/build*/*.bin"]

                [gitlab.artifacts.s3.metrics]
                bucket = "test-bucket"
                patterns = ["**/build*/size.json"]

                [gitlab.artifacts.s3.optional]
                bucket = "test-bucket"
                patterns = ["**/build*/optional.txt"]
                if_clause = 'ENV_VAR_FOO == "foo"'
            """)
        )

        curdir = os.getcwd()
        os.chdir(tmp_path)

        monkeypatch.setenv('IDF_S3_SERVER', 'http://localhost:9100')
        monkeypatch.setenv('IDF_S3_ACCESS_KEY', 'minioadmin')
        monkeypatch.setenv('IDF_S3_SECRET_KEY', 'minioadmin')

        monkeypatch.setenv('IDF_PATH', tmp_dir)

        yield

        os.chdir(curdir)

    @pytest.fixture
    def s3_client(self) -> minio.Minio:
        client = ArtifactManager().s3_client
        assert client is not None

        # Drop and recreate bucket before test
        try:
            for obj in client.list_objects('test-bucket', recursive=True):
                client.remove_object('test-bucket', obj.object_name)

            client.remove_bucket('test-bucket')
        except minio.error.S3Error as e:
            logging.error(f'Error removing bucket: {e}')
            pass
        client.make_bucket('test-bucket')
        return client

    @pytest.mark.parametrize(
        'set_env_var_foo',
        [
            True,
            False,
        ],
    )
    def test_cli_upload_download_artifacts(
        self, s3_client, tmp_path, sample_artifacts_dir, monkeypatch, set_env_var_foo
    ):
        # in this test we use subprocess, since env var is monkeypatched

        # Mock git functions that would be called
        commit_sha = 'cli_test_sha_123'

        if set_env_var_foo:
            monkeypatch.setenv('ENV_VAR_FOO', 'foo')

        # Upload artifacts
        subprocess.run(
            [
                'idf-ci',
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                commit_sha,
                '--type',
                'flash',
            ],
            check=True,
        )
        objs = list(s3_client.list_objects('test-bucket', recursive=True))
        assert len(objs) == 1
        assert objs[0].object_name == f'espressif/esp-idf/{commit_sha}/app/build_esp32_build/test.bin'

        # upload optional
        subprocess.run(
            [
                'idf-ci',
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                commit_sha,
                '--type',
                'optional',
            ],
            check=True,
        )
        objs = list(s3_client.list_objects('test-bucket', recursive=True))
        if set_env_var_foo:
            assert len(objs) == 2
            assert objs[0].object_name == f'espressif/esp-idf/{commit_sha}/app/build_esp32_build/optional.txt'
        else:
            assert len(objs) == 1
            assert objs[0].object_name == f'espressif/esp-idf/{commit_sha}/app/build_esp32_build/test.bin'

        shutil.rmtree(sample_artifacts_dir)

        # download and check if the files were uploaded
        subprocess.run(
            [
                'idf-ci',
                'gitlab',
                'download-artifacts',
                '--commit-sha',
                commit_sha,
                '--type',
                'flash',
                str(tmp_path),
            ],
            check=True,
        )

        assert sorted(os.listdir(sample_artifacts_dir)) == ['test.bin']
        assert open(sample_artifacts_dir / 'test.bin').read() == 'Binary content'

        # generate presigned URL
        presigned_urls = ArtifactManager().generate_presigned_json(
            commit_sha=commit_sha,
            artifact_type='flash',
        )

        # Save presigned URLs to a file
        presigned_json_path = os.path.join(tmp_path, 'presigned.json')
        with open(presigned_json_path, 'w') as f:
            json.dump(presigned_urls, f)

        # Remove S3 credentials
        monkeypatch.delenv('IDF_S3_ACCESS_KEY')

        # Try to download using presigned.json
        subprocess.run(
            [
                'idf-ci',
                'gitlab',
                'download-artifacts',
                '--commit-sha',
                commit_sha,
                '--presigned-json',
                presigned_json_path,
                str(tmp_path),
            ],
            check=True,
        )

        assert os.path.exists(os.path.join(tmp_path, 'app/build_esp32_build/test.bin'))

    def test_cli_generate_presigned_json(self, runner):
        # Mock git functions that would be called
        commit_sha = 'cli_test_sha_123'

        # First upload some artifacts
        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                commit_sha,
                '--type',
                'flash',
            ],
        )
        assert result.exit_code == 0

        # Generate presigned URLs
        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'generate-presigned-json',
                '--commit-sha',
                commit_sha,
                '--type',
                'flash',
            ],
        )
        assert result.exit_code == 0

        # Parse the output JSON
        presigned_urls = json.loads(result.output)
        assert len(presigned_urls) == 1
        assert 'app/build_esp32_build/test.bin' in presigned_urls

        # Verify the presigned URL is valid by downloading the file
        response = requests.get(presigned_urls['app/build_esp32_build/test.bin'])
        assert response.status_code == 200
        assert response.text == 'Binary content'

    def test_download_without_s3_credentials(self, runner, tmp_path, monkeypatch):
        # Remove S3 credentials
        monkeypatch.delenv('IDF_S3_ACCESS_KEY')

        # Try to download artifacts
        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'download-artifacts',
                '--commit-sha',
                'test_sha',
                '--type',
                'flash',
                str(tmp_path),
            ],
        )

        assert result.exit_code != 0
        assert isinstance(result.exception, S3Error)
        assert 'Configure S3 storage to download artifacts' in result.exception.args

    def test_upload_without_s3_credentials(self, runner, tmp_path, monkeypatch):
        # Remove S3 credentials
        monkeypatch.delenv('IDF_S3_ACCESS_KEY')

        # Try to upload artifacts
        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                'test_sha',
                '--type',
                'flash',
                str(tmp_path),
            ],
        )

        assert result.exit_code != 0
        assert isinstance(result.exception, S3Error)
        assert 'Configure S3 storage to upload artifacts' in result.exception.args
